import contextlib

from .model.naming import ReferenceByName
from .model.position import Point, Position
from .validation.validation import Result, Issue, IssueType, IssueSeverity


def unserialize_result(json_result, root_unserializer) -> Result:
    result = Result(root=root_unserializer(json_result['root']) if 'root' in json_result else None)
    result.issues = [unserialize_issue(issue) for issue in json_result['issues']] if 'issues' in json_result else []
    return result


def unserialize_issue(json_issue) -> Issue:
    return Issue(
        type=unserialize_issue_type(json_issue['type']) if json_issue['type'] else None,
        message=json_issue['message'] if json_issue['message'] else None,
        severity=unserialize_issue_severity(json_issue['severity']) if json_issue['severity'] else None,
        position=unserialize_position(json_issue['position']) if json_issue['position'] else None
    )


def unserialize_issue_type(json_issue_type) -> IssueType or None:
    with contextlib.suppress(Exception):
        return IssueType[json_issue_type]


def unserialize_issue_severity(json_issue_severity) -> IssueSeverity:
    with contextlib.suppress(Exception):
        return IssueSeverity[json_issue_severity]


def check_type(json, expected_type):
    if "#type" not in json:
        raise Exception("type not specified, expected %s" % expected_type)
    if json["#type"] != expected_type:
        raise Exception("unexpected type, expected %s but found %s" % (expected_type, json["#type"]))


def unserialize_point(json_point):
    return Point(
        line=json_point['line'] if 'line' in json_point else None,
        column=json_point['column'] if 'column' in json_point else None,
    )


def unserialize_position(json_position):
    return Position(
        start=unserialize_point(json_position['start']) if 'start' in json_position else None,
        end=unserialize_point(json_position['end']) if 'end' in json_position else None
    )


def unserialize_reference_by_name(json_reference_by_name):
    return ReferenceByName(
        name=json_reference_by_name['name'] if 'name' in json_reference_by_name else None
    )


def unserialize_long(json):
    return json



from dataclasses import dataclass, field
from typing import Optional, List, Sequence

from antlr4 import ParserRuleContext, TerminalNode, Token
from antlr4.tree.Tree import ParseTree

from pylasu.model import Origin, Position, Point
from pylasu.model.model import internal_property, Node
from pylasu.model.position import Source
from pylasu.support import extension_method

import inspect


@dataclass
class ParseTreeOrigin(Origin):
    parse_tree: ParseTree
    source: Source = None

    @internal_property
    def position(self) -> Optional[Position]:
        return self.parse_tree.to_position(self.source)

    @internal_property
    def source_text(self) -> Optional[str]:
        return self.parse_tree.get_original_text()


def token_start_point(token: Token):
    return Point(token.line, token.column)


def token_end_point(token: Token):
    if token.type == Token.EOF:
        return token.start_point
    else:
        return token.start_point + token.text


Token.start_point = property(token_start_point)
Token.end_point = property(token_end_point)


@extension_method(ParserRuleContext)
def to_position(self: ParserRuleContext, source: Source = None):
    # In case of an empty input, the start token will be EOF and the end token will be None
    if self.stop and self.start.start_point <= self.stop.end_point:
        return Position(self.start.start_point, self.stop.end_point, source)
    else:
        # In case of parse errors, sometimes ANTLR inserts nodes that end before they start
        return Position(self.start.start_point, self.start.end_point, source)


@extension_method(TerminalNode)
def to_position(self: TerminalNode, source: Source = None):
    return self.symbol.to_position(source)


@extension_method(Token)
def to_position(self: Token, source: Source = None):
    return Position(self.start_point, self.end_point, source)


@extension_method(ParseTree)
def get_original_text(self: ParseTree) -> str:
    return self.getText()


@extension_method(ParserRuleContext)
def get_original_text(self: ParserRuleContext) -> str:
    a = self.start.start
    b = self.stop.stop
    return self.start.getInputStream().getText(a, b)


@extension_method(Node)
def with_parse_tree(self: Node, parse_tree: Optional[ParseTree], source: Source = None):
    """Set the origin of the AST node as a ParseTreeOrigin, providing the parse_tree is not None.
If the parse_tree is None, no operation is performed."""
    if parse_tree:
        self.origin = ParseTreeOrigin(parse_tree=parse_tree, source=source)
    return self


def generate_nodes_classes_for_parser(parser_class: type, ns: dict):
    for name, definition in parse_tree_node_definitions(parser_class):
        fields = {"__annotations__": {}}
        property_names = []
        aliases = {}
        for child_name, value in parse_tree_node_children(definition):
            default_value = None
            field_type = None
            if inspect.isfunction(value):
                n_args = len(inspect.signature(value).parameters)
                if n_args == 2:
                    suffix = "_list"
                    child_name += suffix
                    field_type = List
                    default_value = field(default_factory=list)
                    aliases[child_name] = child_name[:-len(suffix)]
            fields[child_name] = default_value
            fields["__annotations__"][child_name] = field_type
            property_names.append(child_name)
        name = ast_node_name(name)
        class_def = type(name, (Node,), fields)
        class_def.properties = properties_method(property_names)
        class_def.from_parse_tree = from_parse_tree_function(class_def, property_names, aliases, ns)
        ns[name] = dataclass(class_def)


def ast_node_name(parse_tree_node_name):
    if parse_tree_node_name.endswith("Context"):
        parse_tree_node_name = parse_tree_node_name[:-len("Context")]
    return parse_tree_node_name


def properties_method(property_names):
    return internal_property(lambda node: ((p, getattr(node, p)) for p in property_names))


def stop_token(node):
    if isinstance(node, TerminalNode):
        return node.symbol
    elif isinstance(node, ParserRuleContext):
        return node.stop


def start_token(node):
    if isinstance(node, TerminalNode):
        return node.symbol
    elif isinstance(node, ParserRuleContext):
        return node.start


def make_ast_node_or_value(parse_tree_node, prev_node, ns, parent: Node, source: Source):
    ast_node_type_name = ast_node_name(type(parse_tree_node).__name__)
    if ast_node_type_name in ns:
        ast_node = ns[ast_node_type_name].from_parse_tree(parse_tree_node, parent, source)
        return ast_node
    else:
        return parse_tree_node.getText()


def from_parse_tree_function(node_class, property_names, aliases, ns: dict):
    def from_parse_tree(parse_tree: ParseTree, parent: Node = None, source: Source = None):
        node = node_class().with_parent(parent).with_parse_tree(parse_tree, source)
        last_pt_node = parse_tree
        for prop in property_names:
            if prop in aliases:
                pt_prop = aliases[prop]
            else:
                pt_prop = prop
            prop_val = getattr(parse_tree, pt_prop)
            if inspect.ismethod(prop_val):
                prop_val = prop_val()
            if isinstance(prop_val, Sequence) and not isinstance(prop_val, str):
                children = []
                for child in prop_val:
                    children.append(make_ast_node_or_value(child, last_pt_node, ns, node, source))
                    last_pt_node = child
                setattr(node, prop, children)
            elif prop_val is not None:
                setattr(node, prop, make_ast_node_or_value(prop_val, last_pt_node, ns, node, source))
                last_pt_node = prop_val
        return node
    return from_parse_tree


def parse_tree_node_definitions(parser_class: type):
    for name in dir(parser_class):
        definition = getattr(parser_class, name)
        if isinstance(definition, type) and name != "__class__":
            yield name, definition


def parse_tree_node_children(parse_tree_node_class: type):
    for x in dir(parse_tree_node_class):
        if x not in dir(ParserRuleContext) and x != "parser":
            yield x, getattr(parse_tree_node_class, x)



from io import IOBase, BytesIO

from pyecore.ecore import EObject, MetaEClass, EAttribute, EString, EReference
from pyecore.resources import ResourceSet, URI
from pyecore.resources.json import JsonResource as BaseJsonResource

from pylasu import StrumentaLanguageSupport as starlasu

nsURI = "https://strumenta.com/kolasu/transpilation/v1"
name = "StrumentaLanguageSupportTranspilation"


class JsonResource(BaseJsonResource):

    def open_out_stream(self, other=None):
        if isinstance(other, IOBase):
            return other
        else:
            return super().open_out_stream(other)


class TranspilationTrace(EObject, metaclass=MetaEClass):
    # Note: we use camelCase here because Pyecore's JSON serialization doesn't handle having different names for
    # Python attributes and their corresponding Ecore structural features.
    originalCode = EAttribute(eType=EString)
    sourceResult = EReference(containment=True, eType=starlasu.Result)
    targetResult = EReference(containment=True, eType=starlasu.Result)
    generatedCode = EAttribute(eType=EString)
    issues = EReference(containment=True, eType=starlasu.Issue, upper=-1)

    def __init__(self, *, original_code=None, source_result=None, target_result=None, generated_code=None, issues=None):
        super().__init__()
        if original_code is not None:
            self.originalCode = original_code
        if source_result is not None:
            self.sourceResult = source_result
        if target_result is not None:
            self.targetResult = target_result
        if generated_code is not None:
            self.generatedCode = generated_code
        if issues:
            self.issues.extend(issues)

    def save_as_json(self, name, *packages):
        rset = ResourceSet()
        rset.resource_factory['json'] = JsonResource
        resource = rset.create_resource(URI(name))
        for pkg in packages:
            package_resource = rset.create_resource(URI(pkg.nsURI))
            package_resource.contents.append(pkg)
        resource.contents.append(self)
        with BytesIO() as out:
            resource.save(out)
            return out.getvalue().decode('utf-8')



from dataclasses import dataclass
from typing import Optional

from pylasu.model import Position, Node


@dataclass
class ErrorNode:
    """An AST node that marks the presence of an error,
     for example a syntactic or semantic error in the original tree."""
    message: str = None
    position: Optional[Position] = None


@dataclass
class GenericErrorNode(Node, ErrorNode):
    error: Optional[Exception] = None

    def __post_init__(self):
        if not self.message:
            if self.error:
                self.message = f"Exception {self.error}"
            else:
                self.message = "Unspecified error node"



import typing
from enum import EnumMeta, Enum
from typing import Callable


def getannotations(cls):
    import inspect
    try:  # On Python 3.10+
        return inspect.getannotations(cls)
    except AttributeError:
        if isinstance(cls, type):
            return cls.__dict__.get('__annotations__', None)
        else:
            return getattr(cls, '__annotations__', None)


def get_type_origin(tp):
    if hasattr(typing, "get_origin"):
        return typing.get_origin(tp)
    elif hasattr(tp, "__origin__"):
        return tp.__origin__
    elif tp is typing.Generic:
        return typing.Generic
    else:
        return None


def is_enum_type(attr_type):
    return isinstance(attr_type, EnumMeta) and issubclass(attr_type, Enum)


def is_sequence_type(attr_type):
    return isinstance(get_type_origin(attr_type), type) and \
        issubclass(get_type_origin(attr_type), typing.Sequence)


def get_type_arguments(tp):
    if hasattr(typing, "get_args"):
        return typing.get_args(tp)
    elif hasattr(tp, "__args__"):
        res = tp.__args__
        if get_type_origin(tp) is Callable and res[0] is not Ellipsis:
            res = (list(res[:-1]), res[-1])
        return res
    return ()



def extension_method(cls):
    """Installs the decorated function as an extension method on cls.
    See https://mail.python.org/pipermail/python-dev/2008-January/076194.html"""
    def decorator(func):
        name = func.__name__
        if name in cls.__dict__:
            raise Exception(f"{cls} already has a member called {name}")
        setattr(cls, name, func)
        return func
    return decorator




from .StrumentaLanguageSupport import getEClassifier, eClassifiers
from .StrumentaLanguageSupport import name, nsURI, nsPrefix, eClass
from .StrumentaLanguageSupport import BigDecimal, BigInteger, LocalDate, LocalTime, LocalDateTime, Point, Position, Origin, Destination, NodeDestination, TextFileDestination, ASTNode, Statement, Expression, EntityDeclaration, IssueType, IssueSeverity, Issue, PossiblyNamed, Named, ReferenceByName, Result


from . import StrumentaLanguageSupport

__all__ = ['BigDecimal', 'BigInteger', 'LocalDate', 'LocalTime', 'LocalDateTime', 'Point', 'Position', 'Origin', 'Destination', 'NodeDestination', 'TextFileDestination',
           'ASTNode', 'Statement', 'Expression', 'EntityDeclaration', 'IssueType', 'IssueSeverity', 'Issue', 'PossiblyNamed', 'Named', 'ReferenceByName', 'Result']

eSubpackages = []
eSuperPackage = None
StrumentaLanguageSupport.eSubpackages = eSubpackages
StrumentaLanguageSupport.eSuperPackage = eSuperPackage

LocalDateTime.date.eType = LocalDate
LocalDateTime.time.eType = LocalTime
Position.start.eType = Point
Position.end.eType = Point
NodeDestination.node.eType = ASTNode
TextFileDestination.position.eType = Position
ASTNode.position.eType = Position
ASTNode.origin.eType = Origin
ASTNode.destination.eType = Destination
Issue.position.eType = Position
#  TODO eGenericType not supported ReferenceByName.referenced.eType =
ReferenceByName.referenced.eType = ASTNode
#  TODO eGenericType not supported
Result.root.eType = ASTNode
Result.issues.eType = Issue

otherClassifiers = [BigDecimal, BigInteger, IssueType, IssueSeverity]

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)



from typing import Optional, Any

from antlr4 import ParserRuleContext
from antlr4.tree.Tree import ParseTree

from pylasu.model import Node, Origin
from pylasu.parsing.parse_tree import ParseTreeOrigin, with_parse_tree
from pylasu.transformation.transformation import ASTTransformer


class ParseTreeToASTTransformer(ASTTransformer):
    """Implements a transformation from an ANTLR parse tree (the output of the parser) to an AST
    (a higher-level representation of the source code)."""

    def transform(self, source: Optional[Any], parent: Optional[Node] = None) -> Optional[Node]:
        """Performs the transformation of a node and, recursively, its descendants. In addition to the overridden
method, it also assigns the parseTreeNode to the AST node so that it can keep track of its position.
However, a node factory can override the parseTreeNode of the nodes it creates (but not the parent)."""
        node = super().transform(source, parent)
        if node and node.origin and isinstance(source, ParserRuleContext):
            with_parse_tree(node, source)
        return node

    def get_source(self, node: Node, source: Any) -> Any:
        origin = node.origin
        if isinstance(origin, ParseTreeOrigin):
            return origin.parse_tree
        else:
            return source

    def as_origin(self, source: Any) -> Optional[Origin]:
        if isinstance(source, ParseTree):
            return ParseTreeOrigin(source)
        else:
            return None



from .model import Destination, Node, Origin, internal_field, internal_properties
from .naming import Named, PossiblyNamed, ReferenceByName
from .position import Point, Position, pos, Source
from .traversing import walk, walk_ancestors, walk_descendants, walk_leaves_first
from .processing import children, search_by_type



from dataclasses import dataclass, field
from pathlib import Path


@dataclass(order=True)
class Point:
    line: int
    column: int

    def __post_init__(self):
        if self.line < 1:
            raise Exception(f"Line {self.line} cannot be less than 1")
        if self.column < 0:
            raise Exception(f"Column {self.column} cannot be less than 0")

    def is_before(self, other: "Point"):
        return self < other

    def __add__(self, other):
        if isinstance(other, str):
            if len(other) == 0:
                return self
            line = self.line
            column = self.column
            i = 0
            while i < len(other):
                if other[i] == '\n' or other[i] == '\r':
                    line += 1
                    column = 0
                    if other[i] == '\r' and i < len(other) - 1 and other[i + 1] == '\n':
                        i += 1  # Count the \r\n sequence as 1 line
                else:
                    column += 1
                i += 1
            return Point(line, column)
        else:
            raise NotImplementedError()

    def __repr__(self):
        return f"Line {self.line}, Column {self.column}"

    def __str__(self):
        return f"{self.line}:{self.column}"


class Source:
    pass


@dataclass
class SourceSet:
    name: str
    root: Path


@dataclass
class SourceSetElement(Source):
    sourceSet: SourceSet
    relativePath: Path


@dataclass
class FileSource(Source):
    file: Path

    def __str__(self):
        return str(self.file)


@dataclass
class StringSource(Source):
    code: str = None


@dataclass
class URLSource(Source):
    url: str = None


@dataclass(order=True)
class Position:
    """An area in a source file, from start to end.
    The start point is the point right before the starting character.
    The end point is the point right after the last character.
    An empty position will have coinciding points.

    Consider a file with one line, containing the text "HELLO".
    The Position of such text will be Position(Point(1, 0), Point(1, 5))."""
    start: Point
    end: Point
    source: Source = field(compare=False, default=None)

    def __post_init__(self):
        if self.end < self.start:
            raise Exception(f"End point can't be before starting point: {self.start} – {self.end}")

    def __contains__(self, pos):
        return isinstance(pos, Position) and self.start <= pos.start and self.end >= pos.end

    def __repr__(self):
        return f"Position(start={self.start}, end={self.end}"\
               + (f", source={self.source}" if self.source is not None else "")

    def __str__(self):
        str_rep = (f"{self.source}:" if self.source is not None else "") + str(self.start)
        if self.start != self.end:
            str_rep += f"-{self.end}"
        return str_rep


def pos(start_line: int, start_col: int, end_line: int, end_col: int, source: Source = None):
    """Utility function to create a Position"""
    return Position(Point(start_line, start_col), Point(end_line, end_col), source)



import enum
from dataclasses import dataclass, field
from typing import List

from pylasu.model import Position, Node


class IssueType(enum.Enum):
    LEXICAL = 0
    SYNTACTIC = 1
    SEMANTIC = 2


class IssueSeverity(enum.Enum):
    ERROR = 30
    WARNING = 20
    INFO = 10


@dataclass
class Issue:
    type: IssueType
    message: str
    severity: IssueSeverity = IssueSeverity.ERROR
    position: Position = None

    def __str__(self):
        msg = f"{self.severity.name.capitalize()} ({self.type.name.lower()}): {self.message}"
        if self.position:
            msg += f" @ {self.position}"
        return msg

    @staticmethod
    def semantic(message: str, severity: IssueSeverity = IssueSeverity.ERROR, position: Position = None):
        return Issue(IssueType.SEMANTIC, message, severity, position)


@dataclass
class WithIssues:
    """Many classes have the necessity of tracking issues"""
    issues: List[Issue] = field(default_factory=list, init=False)


@dataclass
class Result(WithIssues):
    root: Node



from collections.abc import Iterable
from typing import Callable, List, Set

from . import walk
from .model import Node, internal_property
from ..support import extension_method


@extension_method(Node)
def assign_parents(self: Node):
    """Sets or corrects the parent of all AST nodes.

    Pylasu does not see set/add/delete operations on the AST nodes, so this function should be called manually after
    modifying the AST, unless you've taken care of assigning the parents yourself.

    :param self: the root of the AST subtree to start from.
    """
    for node in children(self):
        node.parent = self
        assign_parents(node)


def children(self: Node):
    yield from nodes_in(p.value for p in self.properties)


Node.children = internal_property(children)


def nodes_in(iterable):
    for value in iterable:
        if isinstance(value, Node):
            yield value
        elif isinstance(value, Iterable) and not isinstance(value, str):
            yield from nodes_in(value)


@extension_method(Node)
def search_by_type(self: Node, target_type, walker=walk):
    for node in walker(self):
        if isinstance(node, target_type):
            yield node


@extension_method(Node)
def transform_children(self: Node, operation: Callable[[Node], Node]):
    for prop in self.properties:
        name = prop.name
        value = prop.value
        if isinstance(value, Node):
            new_value = operation(value)
            if new_value != value:
                setattr(self, name, new_value)
        elif isinstance(value, List):
            setattr(self, name, [operation(item) if isinstance(item, Node) else item for item in value])
        elif isinstance(value, Set):
            raise Exception("Sets are not supported currently")


@extension_method(Node)
def replace_with(self: Node, other: Node):
    """Replace this node with another (by modifying the children of the parent node).
    For this to work, this node must have a parent assigned.

    :param self: the node to replace.
    :param other: the replacement node."""
    if not self.parent:
        raise Exception("Parent not set, cannot replace node")
    transform_children(self.parent, lambda x: other if x == self else x)



import inspect
from abc import ABC, abstractmethod, ABCMeta
from dataclasses import Field, MISSING, dataclass, field
from typing import Optional, Callable, List

from .position import Position, Source
from .reflection import Multiplicity, PropertyDescription
from ..reflection import getannotations, get_type_arguments, is_sequence_type


class internal_property(property):
    pass


def internal_properties(*props: str):
    def decorate(cls: type):
        cls.__internal_properties__ = [*Node.__internal_properties__, *props]
        return cls
    return decorate


class InternalField(Field):
    pass


def internal_field(
        *, default=MISSING, default_factory=MISSING, init=True, repr=True, hash=None, compare=True, metadata=None,
        kw_only=False):
    """Return an object to identify internal dataclass fields. The arguments are the same as dataclasses.field."""

    if default is not MISSING and default_factory is not MISSING:
        raise ValueError('cannot specify both default and default_factory')
    try:
        # Python 3.10+
        return InternalField(default, default_factory, init, repr, hash, compare, metadata, kw_only)
    except TypeError:
        return InternalField(default, default_factory, init, repr, hash, compare, metadata)


class Origin(ABC):
    @internal_property
    @abstractmethod
    def position(self) -> Optional[Position]:
        pass

    @internal_property
    def source_text(self) -> Optional[str]:
        return None

    @internal_property
    def source(self) -> Optional[Source]:
        return self.position.source if self.position is not None else None


@dataclass
class CompositeOrigin(Origin):
    elements: List[Origin] = field(default_factory=list)
    position: Optional[Position] = None
    source_text: Optional[str] = None


class Destination(ABC):
    pass


@dataclass
class CompositeDestination(Destination):
    elements: List[Destination] = field(default_factory=list)


@dataclass
class TextFileDestination(Destination):
    position: Optional[Position] = None


def is_internal_property_or_method(value):
    return isinstance(value, internal_property) or isinstance(value, InternalField) or isinstance(value, Callable)


def provides_nodes(decl_type):
    return isinstance(decl_type, type) and issubclass(decl_type, Node)


class Concept(ABCMeta):

    def __init__(cls, what, bases=None, dict=None):
        super().__init__(what, bases, dict)
        cls.__internal_properties__ = ["origin", "destination", "parent", "position", "position_override"]

    @property
    def node_properties(cls):
        names = set()
        for cl in cls.__mro__:
            yield from cls._direct_node_properties(cl, names)

    def _direct_node_properties(cls, cl, known_property_names):
        anns = getannotations(cl)
        if not anns:
            return
        for name in anns:
            if name not in known_property_names and cls.is_node_property(name):
                is_child_property = False
                multiplicity = Multiplicity.SINGULAR
                if name in anns:
                    decl_type = anns[name]
                    if is_sequence_type(decl_type):
                        multiplicity = Multiplicity.MANY
                        type_args = get_type_arguments(decl_type)
                        if len(type_args) == 1:
                            is_child_property = provides_nodes(type_args[0])
                    else:
                        is_child_property = provides_nodes(decl_type)
                known_property_names.add(name)
                yield PropertyDescription(name, is_child_property, multiplicity)
        for name in dir(cl):
            if name not in known_property_names and cls.is_node_property(name):
                known_property_names.add(name)
                yield PropertyDescription(name, False)

    def is_node_property(cls, name):
        return not name.startswith('_') \
            and name not in cls.__internal_properties__ \
            and name not in [n for n, v in inspect.getmembers(cls, is_internal_property_or_method)]


class Node(Origin, Destination, metaclass=Concept):
    origin: Optional[Origin] = None
    destination: Optional[Destination] = None
    parent: Optional["Node"] = None
    position_override: Optional[Position] = None

    def __init__(self, origin: Optional[Origin] = None, parent: Optional["Node"] = None,
                 position_override: Optional[Position] = None):
        self.origin = origin
        self.parent = parent
        self.position_override = position_override

    def with_origin(self, origin: Optional[Origin]):
        self.origin = origin
        return self

    def with_parent(self, parent: Optional["Node"]):
        self.parent = parent
        return self

    def with_position(self, position: Optional[Position]):
        self.position = position
        return self

    @internal_property
    def position(self) -> Optional[Position]:
        return self.position_override if self.position_override is not None\
            else self.origin.position if self.origin is not None else None

    @position.setter
    def position(self, position: Optional[Position]):
        self.position_override = position

    @internal_property
    def source_text(self) -> Optional[str]:
        return self.origin.source_text if self.origin is not None else None

    @internal_property
    def source(self) -> Optional[Source]:
        return self.origin.source if self.origin is not None else None

    @internal_property
    def properties(self):
        return (PropertyDescription(p.name, p.provides_nodes, p.multiplicity, getattr(self, p.name))
                for p in self.__class__.node_properties)

    @internal_property
    def _fields(self):
        yield from (name for name, _ in self.properties)

    @internal_property
    def node_type(self):
        return type(self)


def concept_of(node):
    properties = dir(node)
    if "__concept__" in properties:
        node_type = node.__concept__
    elif "node_type" in properties:
        node_type = node.node_type
    else:
        node_type = type(node)
    if isinstance(node_type, Concept):
        return node_type
    else:
        raise Exception(f"Not a concept: {node_type} of {node}")



import unittest

from pylasu.model import Node


def assert_asts_are_equal(
        case: unittest.TestCase,
        expected: Node, actual: Node,
        context: str = "<root>", consider_position: bool = False
):
    if expected.node_type != actual.node_type:
        case.fail(f"{context}: expected node of type {expected.node_type}, "
                  f"but found {actual.node_type}")
    if consider_position:
        case.assertEqual(expected.position, actual.position, f"{context}.position")
    for expected_property in expected.properties:
        try:
            actual_property = next(filter(lambda p: p.name == expected_property.name, actual.properties))
        except StopIteration:
            case.fail(f"No property {expected_property.name} found at {context}")
        actual_prop_value = actual_property.value
        expected_prop_value = expected_property.value
        if expected_property.provides_nodes:
            if expected_property.multiple:
                assert_multi_properties_are_equal(
                    case, expected_property, expected_prop_value, actual_prop_value, context, consider_position)
            else:
                assert_single_properties_are_equal(case, expected_property, expected_prop_value, actual_prop_value,
                                                   context, consider_position)
        # TODO not yet supported elif expected_property.property_type == PropertyType.REFERENCE:
        else:
            case.assertEqual(
                expected_prop_value, actual_prop_value,
                f"{context}, comparing property {expected_property.name} of {expected.node_type}")


def assert_single_properties_are_equal(case, expected_property, expected_prop_value, actual_prop_value, context,
                                       consider_position):
    if expected_prop_value is None and actual_prop_value is not None:
        case.assertEqual(expected_prop_value, actual_prop_value,
                         f"{context}.{expected_property.name}")
    elif expected_prop_value is not None and actual_prop_value is None:
        case.assertEqual(expected_prop_value, actual_prop_value,
                         f"{context}.{expected_property.name}")
    elif expected_prop_value is None and actual_prop_value is None:
        # that is ok
        pass
    else:
        case.assertIsInstance(actual_prop_value, Node)
        assert_asts_are_equal(
            case, expected_prop_value, actual_prop_value,
            context=f"{context}.{expected_property.name}",
            consider_position=consider_position)


def assert_multi_properties_are_equal(case, expected_property, expected_prop_value, actual_prop_value, context,
                                      consider_position):
    # TODO IgnoreChildren
    case.assertEqual(actual_prop_value is None, expected_prop_value is None,
                     f"{context}.{expected_property.name} nullness")
    if actual_prop_value is not None and expected_prop_value is not None:
        case.assertEqual(len(actual_prop_value), len(expected_prop_value),
                         f"{context}.{expected_property.name} length")
    for expected_it, actual_it, i in \
            zip(expected_prop_value, actual_prop_value, range(len(expected_prop_value))):
        assert_asts_are_equal(case, expected_it, actual_it, f"{context}[{i}]",
                              consider_position=consider_position)



from . import Position
from .model import Node
from ..support import extension_method


@extension_method(Node)
def walk(self: Node):
    """Walks the whole AST starting from this node, depth-first."""
    yield self
    for child in self.children:
        yield from walk(child)


@extension_method(Node)
def walk_within(self: Node, position: Position):
    """Walks the AST within the given [position] starting from this node, depth-first.
    :param self: the node from which to start the walk.
    :param position: the position within which the walk should remain."""
    if self.position in position:
        yield self
        for child in self.children:
            yield from walk_within(child, position)
    elif position in self.position:
        for child in self.children:
            yield from walk_within(child, position)


@extension_method(Node)
def walk_leaves_first(self: Node):
    """Performs a post-order (or leaves-first) node traversal starting with a given node."""
    for child in self.children:
        yield from walk_leaves_first(child)
    yield self


@extension_method(Node)
def walk_ancestors(self: Node):
    """Iterator over the sequence of nodes from this node's parent all the way up to the root node."""
    if self.parent is not None:
        yield self.parent
        yield from walk_ancestors(self.parent)


@extension_method(Node)
def walk_descendants(self: Node, walker=walk, restrict_to=Node):
    """Walks the whole AST starting from the child nodes of this node.

    :param self: the node from which to start the walk, which is NOT included in the walk.
    :param walker: a function that generates a sequence of nodes. By default this is the depth-first "walk" method.
    For post-order traversal, use "walk_leaves_first".
    :param restrict_to: optional type filter. By default, all nodes (i.e., subclasses of Node) are included, but you can
    limit the walk to only a subtype of Node.
    """
    for node in walker(self):
        if node != self and isinstance(node, restrict_to):
            yield node


@extension_method(Node)
def find_ancestor_of_type(self: Node, target: type):
    """Returns the nearest ancestor of this node that is an instance of the target type.

    Note that type is not strictly forced to be a subtype of Node. This is intended to support trait types like
    `Statement` or `Expression`. However, the returned value is guaranteed to be a Node, as only Node instances can be
    part of the hierarchy.

    :param self: the node from which to start the search.
    :param target: the target type.
    """
    for node in walk_ancestors(self):
        if isinstance(node, target):
            return node



from dataclasses import dataclass, field
from io import BytesIO
from typing import List

from pyecore.resources import Resource, ResourceSet, URI

from pylasu import StrumentaLanguageSupport as starlasu
from pylasu.playground.transpilation_trace_ecore import TranspilationTrace as ETranspilationTrace, JsonResource
from pylasu.validation.validation import Result, Issue


@dataclass
class TranspilationTrace:
    original_code: str
    source_result: Result
    target_result: Result
    generated_code: str
    issues: List[Issue] = field(default_factory=list)

    def to_eobject(self, resource: Resource):
        mappings = {}
        return ETranspilationTrace(
            original_code=self.original_code,
            source_result=starlasu.Result(root=self.source_result.root.to_eobject(resource, mappings)),
            target_result=starlasu.Result(root=self.target_result.root.to_eobject(resource, mappings)),
            generated_code=self.generated_code
        )

    def save_as_json(self, name, *packages):
        rset = ResourceSet()
        rset.resource_factory['json'] = JsonResource
        resource = rset.create_resource(URI(name))
        for pkg in packages:
            package_resource = rset.create_resource(URI(pkg.nsURI))
            package_resource.contents.append(pkg)
        resource.contents.append(self.to_eobject(resource))
        with BytesIO() as out:
            resource.save(out)
            return out.getvalue().decode('utf-8')



import enum
from dataclasses import dataclass


class Multiplicity(enum.Enum):
    OPTIONAL = 0
    SINGULAR = 1
    MANY = 2


@dataclass
class PropertyDescription:
    name: str
    provides_nodes: bool
    multiplicity: Multiplicity = Multiplicity.SINGULAR
    value: object = None

    @property
    def multiple(self):
        return self.multiplicity == Multiplicity.MANY



from dataclasses import dataclass

from pylasu.model import Node


@dataclass
class GenericNode(Node):
    """A generic AST node. We use it to represent parts of a source tree that we don't know how to translate yet."""
    parent: Node = None



from dataclasses import dataclass, field
from typing import TypeVar, Generic, Optional, List, Dict


@dataclass
class PossiblyNamed:
    name: str

    def __init__(self, name: str = None):
        self.name = name


@dataclass
class Named(PossiblyNamed):
    name: str


T = TypeVar("T", bound=PossiblyNamed)


@dataclass
class ReferenceByName(Generic[T]):
    name: str
    referred: Optional[T] = field(default=None)

    def __str__(self):
        status = 'Solved' if self.resolved() else 'Unsolved'
        return f"Ref({self.name})[{status}]"

    def __hash__(self):
        return self.name.__hash__() * (7 + 2 if self.resolved() else 1)

    def resolved(self):
        return self.referred is not None

    def resolve(self, scope: 'Scope') -> bool:
        self.referred = scope.lookup(symbol_name=self.name)
        return self.resolved()

    def try_to_resolve(self, candidates: List[T], case_insensitive: bool = False) -> bool:
        """
        Try to resolve the reference by finding a named element with a matching name.
        The name match is performed in a case sensitive or insensitive way depending on the value of caseInsensitive.
        """

        def check_name(candidate: T) -> bool:
            return candidate.name is not None and (
                candidate.name == self.name if not case_insensitive else candidate.name.lower() == self.name.lower())

        self.referred = next((candidate for candidate in candidates if check_name(candidate)), None)
        return self.resolved()


@dataclass
class Symbol(PossiblyNamed):
    pass


@dataclass
class Scope:
    symbols: Dict[str, List[Symbol]] = field(default_factory=list)
    parent: Optional['Scope'] = field(default=None)

    def lookup(self, symbol_name: str, symbol_type: type = Symbol) -> Optional[Symbol]:
        return next((symbol for symbol in self.symbols.get(symbol_name, []) if isinstance(symbol, symbol_type)),
                    self.parent.lookup(symbol_name, symbol_type) if self.parent is not None else None)

    def add(self, symbol: Symbol):
        self.symbols[symbol.name] = self.symbols.get(symbol.name, []) + [symbol]



import types
import typing
from dataclasses import is_dataclass, fields

from pyecore.ecore import EAttribute, ECollection, EObject, EPackage, EReference, MetaEClass, EBoolean, EString, EInt, \
    EEnum
from pyecore.resources import Resource
from pylasu import StrumentaLanguageSupport as starlasu
from pylasu.StrumentaLanguageSupport import ASTNode
from pylasu.emf.model import find_eclassifier
from pylasu.model import Node
from pylasu.model.model import InternalField
from pylasu.reflection import getannotations
from pylasu.reflection.reflection import get_type_origin, is_enum_type, is_sequence_type, get_type_arguments


def resolve_bases(bases):
    """Resolve MRO entries dynamically as specified by PEP 560."""
    if hasattr(types, "resolve_bases"):
        return getattr(types, "resolve_bases")(bases)
    new_bases = list(bases)
    updated = False
    shift = 0
    for i, base in enumerate(bases):
        if isinstance(base, type):
            continue
        if not hasattr(base, "__mro_entries__"):
            continue
        new_base = base.__mro_entries__(bases)
        updated = True
        if not isinstance(new_base, tuple):
            raise TypeError("__mro_entries__ must return a tuple")
        else:
            new_bases[i + shift:i + shift + 1] = new_base
            shift += len(new_base) - 1
    if not updated:
        return bases
    return tuple(new_bases)


class MetamodelBuilder:
    def __init__(self, package_name: str, ns_uri: str, ns_prefix: str = None, resource: Resource = None,
                 base_node_class: type = Node):
        self.package = EPackage(package_name, ns_uri, ns_prefix)
        if resource:
            resource.append(self.package)
        self.data_types = {
            bool: EBoolean,
            int: EInt,
            str: EString,
        }
        self.base_node_class = base_node_class
        self.forward_references = []

    def can_provide_class(self, cls: type):
        return cls.__module__ == self.package.name

    def provide_class(self, cls: type):
        if cls == self.base_node_class:
            return ASTNode
        if not self.can_provide_class(cls):
            if self.package.eResource:
                eclass = find_eclassifier(self.package.eResource, cls)
                if eclass:
                    return eclass
            raise Exception(self.package.name + " cannot provide class " + str(cls))
        eclass = self.package.getEClassifier(cls.__name__)
        if not eclass:
            nmspc = self.setup_attributes(cls)
            bases = self.setup_base_classes(cls)
            eclass = MetaEClass(cls.__name__, resolve_bases(tuple(bases)), nmspc)
            eclass.eClass.ePackage = self.package
            for (type_name, ref) in self.forward_references:
                if type_name == cls.__name__:
                    ref.eType = eclass
            self.forward_references = [(t, r) for t, r in self.forward_references if not r.eType]
        return eclass

    def setup_base_classes(self, cls):
        bases = []
        for c in cls.__mro__[1:]:
            if c == self.base_node_class:
                bases.append(ASTNode)
            elif self.can_provide_class(c):
                bases.append(self.provide_class(c))
            elif self.package.eResource:
                esuperclass = find_eclassifier(self.package.eResource, c)
                if esuperclass:
                    bases.append(esuperclass)
        bases.append(EObject)
        return bases

    def setup_attributes(self, cls):
        anns = getannotations(cls)
        nmspc = {
            "position": EReference("position", starlasu.Position, containment=True)
        }
        for attr in anns if anns else []:
            if attr.startswith('_'):
                continue
            elif is_dataclass(cls):
                field = next((f for f in fields(cls) if f.name == attr), None)
                if isinstance(field, InternalField):
                    continue
            attr_type = anns[attr]
            nmspc[attr] = self.to_structural_feature(attr, attr_type)
        return nmspc

    def to_structural_feature(self, attr, attr_type, unsupported_type_handler=None):  # noqa: C901
        def raise_on_unsupported_type(attr_type, attr):
            raise Exception("Unsupported type " + str(attr_type) + " for attribute " + attr)

        def default_unsupported_type(_, __):
            return EObject

        if not unsupported_type_handler:
            unsupported_type_handler = raise_on_unsupported_type
        if isinstance(attr_type, str):
            return self.to_reference(attr, attr_type)
        elif attr_type in self.data_types:
            return EAttribute(attr, self.data_types[attr_type])
        elif attr_type == object:
            return EAttribute(attr)
        elif self.is_node_type(attr_type):
            return EReference(attr, self.provide_class(attr_type), containment=True)
        elif is_sequence_type(attr_type):
            return self.to_list_reference(attr, attr_type, default_unsupported_type)
        elif get_type_origin(attr_type) == typing.Union:
            return EReference(attr, EObject, containment=True)  # TODO here we could refine the type better
        elif is_enum_type(attr_type):
            return self.to_enum_attribute(attr, attr_type)
        else:
            return unsupported_type_handler(attr_type, attr)

    def is_node_type(self, attr_type):
        return isinstance(attr_type, type) and issubclass(attr_type, self.base_node_class)

    def to_enum_attribute(self, attr, attr_type):
        tp = EEnum(name=attr_type.__name__, literals=attr_type.__members__)
        tp.ePackage = self.package
        self.data_types[attr_type] = tp
        return EAttribute(attr, tp)

    def to_list_reference(self, attr, attr_type, default_unsupported_type):
        type_args = get_type_arguments(attr_type)
        if type_args and len(type_args) == 1:
            ft = self.to_structural_feature(attr, type_args[0], default_unsupported_type)
            ft.upperBound = -1
            return ft
        else:
            raise "Unsupported list type: " + str(attr_type)

    def to_reference(self, attr, attr_type):
        resolved = self.package.getEClassifier(attr_type)
        if resolved:
            return EReference(attr, resolved, containment=True)
        else:
            forward_reference = EReference(attr, containment=True)
            self.forward_references.append((attr_type, forward_reference))
            return forward_reference

    def generate(self):
        if self.forward_references:
            raise Exception("The following classes are missing from " + self.package.name + ": "
                            + ", ".join(n for n, _ in self.forward_references))
        return self.package


# Monkey patch until fix
update_opposite = ECollection._update_opposite


def update_opposite_if_not_none(self, owner, new_value, remove=False):
    if owner:
        update_opposite(self, owner, new_value, remove)


ECollection._update_opposite = update_opposite_if_not_none



import functools
from dataclasses import dataclass, field
from inspect import signature
from typing import Any, Dict, Callable, TypeVar, Generic, Optional, List, Set, Iterable, Type, Union

from pylasu.model import Node, Origin
from pylasu.model.errors import GenericErrorNode
from pylasu.model.model import concept_of
from pylasu.model.reflection import PropertyDescription
from pylasu.transformation.generic_nodes import GenericNode
from pylasu.validation import Issue, IssueSeverity

Child = TypeVar('Child')
Output = TypeVar('Output', bound=Node)
Source = TypeVar('Source')
Target = TypeVar('Target')
node_factory_constructor_type = Callable[[Source, "ASTTransformer", "NodeFactory[Source, Output]"], List[Output]]
node_factory_single_constructor_type = Callable[[Source, "ASTTransformer", "NodeFactory[Source, Output]"], Output]


@dataclass
class PropertyRef:
    name: str

    def get(self, node: Node):
        return getattr(node, self.name)

    def set(self, node: Node, value):
        return setattr(node, self.name, value)


@dataclass
class NodeFactory(Generic[Source, Output]):
    constructor: node_factory_constructor_type
    children: Dict[str, "ChildNodeFactory[Source, Any, Any]"] = field(default_factory=dict)
    finalizer: Callable[[Source], None] = field(default=lambda _: None)

    def with_child(
            self,
            setter: Union[Callable[[Target, Optional[Child]], None], PropertyRef],
            getter: Union[Callable[[Source], Optional[Any]], PropertyRef],
            name: Optional[str] = None,
            target_type: Optional[type] = None
    ) -> "NodeFactory[Source, Output]":
        if not name:
            name = setter.name
        if target_type:
            prefix = target_type.__qualname__ + "#"
        else:
            prefix = ""
        if isinstance(getter, PropertyRef):
            getter = getter.get
        if isinstance(setter, PropertyRef):
            setter = setter.set
        self.children[prefix + name] = ChildNodeFactory(prefix + name, getter, setter)
        return self


@dataclass
class ChildNodeFactory(Generic[Source, Target, Child]):
    name: str
    get: Callable[[Source], Optional[Any]]
    setter: Callable[[Target, Optional[Child]], None]

    def set(self, node: Target, child: Optional[Child]):
        try:
            self.setter(node, child)
        except Exception as e:
            raise Exception(f"{self.name} could not set child {child} of {node} using {self.setter}") from e


"""Sentinel value used to represent the information that a given property is not a child node."""
NO_CHILD_NODE = ChildNodeFactory("", lambda x: x, lambda _, __: None)


class ASTTransformer:
    issues: List[Issue] = []
    "Additional issues found during the transformation process."
    allow_generic_node: bool = True
    factories: Dict[type, NodeFactory]
    "Factories that map from source tree node to target tree node."
    known_classes: Dict[str, Set[type]]

    def __init__(self, issues: List[Issue] = None, allow_generic_node: bool = True):
        self.issues = issues or []
        self.allow_generic_node = allow_generic_node
        self.factories = dict()
        self.known_classes = dict()

    def transform(self, source: Optional[Any], parent: Optional[Node] = None) -> Optional[Node]:
        result = self.transform_into_nodes(source, parent)
        if len(result) == 0:
            return None
        elif len(result) == 1:
            return result[0]
        else:
            raise Exception(f"Cannot transform {source} into a single Node as multiple nodes where produced")

    def transform_into_nodes(self, source: Optional[Any], parent: Optional[Node] = None) -> List[Node]:
        if source is None:
            return []
        elif isinstance(source, Iterable):
            raise Exception(f"Mapping error: received collection when value was expected: {source}")
        factory = self.get_node_factory(type(source))
        if factory:
            nodes = self.make_nodes(factory, source)
            for node in nodes:
                for pd in concept_of(node).node_properties:
                    self.process_child(source, node, pd, factory)
                factory.finalizer(node)
                node.parent = parent

        else:
            if self.allow_generic_node:
                origin = self.as_origin(source)
                nodes = [GenericNode(parent).with_origin(origin)]
                self.issues.append(
                    Issue.semantic(
                        f"Source node not mapped: {type(source).__qualname__}",
                        IssueSeverity.INFO,
                        origin.position if origin else None))
            else:
                raise Exception(f"Unable to transform node {source} (${type(source)})")
        return nodes

    def process_child(self, source, node, pd, factory):
        child_key = type(node).__qualname__ + "#" + pd.name
        if child_key in factory.children:
            child_node_factory = factory.children[child_key]
        elif pd.name in factory.children:
            child_node_factory = factory.children[pd.name]
        else:
            child_node_factory = None
        if child_node_factory:
            if child_node_factory != NO_CHILD_NODE:
                self.set_child(child_node_factory, source, node, pd)
        else:
            # TODO should we support @Mapped / dot-notation?
            factory.children[child_key] = NO_CHILD_NODE

    def as_origin(self, source: Any) -> Optional[Origin]:
        return source if isinstance(source, Origin) else None

    def set_child(self, child_node_factory: ChildNodeFactory, source: Any, node: Node, pd: PropertyDescription):
        src = child_node_factory.get(self.get_source(node, source))
        if pd.multiple:
            child = []
            for child_src in src:
                child.extend(self.transform_into_nodes(child_src, node))
        else:
            child = self.transform(src, node)
        try:
            child_node_factory.set(node, child)
        except Exception as e:
            raise Exception(f"Could not set child {child_node_factory}") from e

    def get_source(self, node: Node, source: Any) -> Any:
        return source

    def make_nodes(self, factory: NodeFactory[Source, Target], source: Source) -> List[Node]:
        try:
            nodes = factory.constructor(source, self, factory)
            for node in nodes:
                if node.origin is None:
                    node.with_origin(self.as_origin(source))
            return nodes
        except Exception as e:
            if self.allow_generic_node:
                return [GenericErrorNode(error=e).with_origin(self.as_origin(source))]
            else:
                raise e

    def get_node_factory(self, node_type: Type[Source]) -> Optional[NodeFactory[Source, Target]]:
        if node_type in self.factories:
            return self.factories[node_type]
        else:
            for superclass in node_type.__mro__[1:]:
                factory = self.get_node_factory(superclass)
                if factory:
                    return factory

    def register_node_factory(
            self, source: Type[Source],
            factory: Union[node_factory_constructor_type, node_factory_single_constructor_type, Type[Target]]
    ) -> NodeFactory[Source, Target]:
        if isinstance(factory, type):
            node_factory = NodeFactory(lambda _, __, ___: [factory()])
        else:
            node_factory = NodeFactory(get_node_constructor_wrapper(factory))
        self.factories[source] = node_factory
        return node_factory

    def register_identity_transformation(self, node_class: Type[Target]):
        self.register_node_factory(node_class, lambda node: node)


def ensure_list(obj):
    if isinstance(obj, list):
        return obj
    elif obj is not None:
        return [obj]
    else:
        return []


def get_node_constructor_wrapper(decorated_function):  # noqa C901
    try:
        sig = signature(decorated_function)
        try:
            sig.bind(1, 2, 3)

            def wrapper(node: Node, transformer: ASTTransformer, factory):
                return ensure_list(decorated_function(node, transformer, factory))
        except TypeError:
            try:
                sig.bind(1, 2)

                def wrapper(node: Node, transformer: ASTTransformer, _):
                    return ensure_list(decorated_function(node, transformer))
            except TypeError:
                sig.bind(1)

                def wrapper(node: Node, _, __):
                    return ensure_list(decorated_function(node))
    except ValueError:
        def wrapper(node: Node, transformer: ASTTransformer, factory):
            return ensure_list(decorated_function(node, transformer, factory))

    functools.update_wrapper(wrapper, decorated_function)
    return wrapper


def ast_transformer(
        node_class: Type[Node],
        transformer: ASTTransformer,
        method_name: str = None):
    """Decorator to register a function as an AST transformer"""
    def decorator(decorated_function):
        if method_name:
            def transformer_method(self, parent: Optional[Node] = None, transformer: ASTTransformer = transformer):
                return transformer.transform(self, parent)
            setattr(node_class, method_name, transformer_method)
        if transformer:
            return transformer.register_node_factory(node_class, decorated_function).constructor
        else:
            return get_node_constructor_wrapper(decorated_function)

    return decorator



from dataclasses import dataclass
from typing import List

from antlr4 import ParserRuleContext, Token

from pylasu.model import Source
from pylasu.validation.validation import WithIssues, IssueType, Issue


@dataclass
class FirstStageResult(WithIssues):
    parse_tree: ParserRuleContext


@dataclass
class LexingResult(WithIssues):
    tokens: List[Token]


@dataclass
class IssuesErrorListener:
    """This Error Listener should be used with ANTLR lexers and parsers to capture issues"""
    type: IssueType
    source: Source
    issues: WithIssues

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        self.issues.append(Issue(type=self.type, message=msg))

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        pass

    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        pass

    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        pass



import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, "pylasu", '__version__.py')) as f:
    exec(f.read(), about)

setup(
    packages=find_packages(exclude=["tests"]),
    version=about['__version__'],
    install_requires=[],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
)



from enum import Enum

from pyecore.ecore import EPackage
from pyecore.resources import Resource

from pylasu.model import Node
from pylasu.support import extension_method


def find_eclassifier_in_resource(cls: type, resource: Resource):
    pkg_name = cls.__module__
    for p in resource.contents:
        if isinstance(p, EPackage) and p.name == pkg_name:
            return p.getEClassifier(cls.__name__)


@extension_method(Resource)
def find_eclassifier(self: Resource, cls: type):
    eclass = find_eclassifier_in_resource(cls, self)
    if not eclass:
        for uri in (self.resource_set.resources if self.resource_set else {}):
            resource = self.resource_set.resources[uri]
            if resource != self:
                eclass = find_eclassifier_in_resource(cls, resource)
                if eclass:
                    return eclass
    return eclass


@extension_method(Node)
def to_eobject(self: Node, resource: Resource, mappings=None):
    if self is None:
        return None
    if mappings is None:
        mappings = {}
    elif id(self) in mappings:
        return mappings[id(self)]
    eclass = resource.find_eclassifier(type(self))
    if not eclass:
        raise Exception("Unknown classifier for " + str(type(self)))
    eobject = eclass()
    mappings[id(self)] = eobject
    for p in self.properties:
        v = p.value
        ev = translate_value(v, resource, mappings)
        if isinstance(v, list):
            eobject.eGet(p.name).extend(ev)
        else:
            eobject.eSet(p.name, ev)
    return eobject


def translate_value(v, resource, mappings):
    if isinstance(v, Enum):
        enum_type = resource.find_eclassifier(type(v))
        if enum_type:
            return enum_type.getEEnumLiteral(v.name)
        else:
            raise Exception("Unknown enum " + str(type(v)))
    if isinstance(v, list):
        return [translate_value(x, resource, mappings) for x in v]
    elif isinstance(v, Node):
        return to_eobject(v, resource, mappings)
    else:
        return v



"""Definition of meta model 'StrumentaLanguageSupport'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'StrumentaLanguageSupport'
nsURI = 'https://strumenta.com/kolasu/v2'
nsPrefix = ''

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)
IssueType = EEnum('IssueType', literals=['LEXICAL', 'SYNTACTIC', 'SEMANTIC'])

IssueSeverity = EEnum('IssueSeverity', literals=['ERROR', 'WARNING', 'INFO'])


BigDecimal = EDataType('BigDecimal', instanceClassName='java.math.BigDecimal')

BigInteger = EDataType('BigInteger', instanceClassName='java.math.BigInteger')


class LocalDate(EObject, metaclass=MetaEClass):

    year = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)
    month = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)
    dayOfMonth = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)

    def __init__(self, *, year=None, month=None, dayOfMonth=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if year is not None:
            self.year = year

        if month is not None:
            self.month = month

        if dayOfMonth is not None:
            self.dayOfMonth = dayOfMonth


class LocalTime(EObject, metaclass=MetaEClass):

    hour = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)
    minute = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)
    second = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)
    nanosecond = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)

    def __init__(self, *, hour=None, minute=None, second=None, nanosecond=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if hour is not None:
            self.hour = hour

        if minute is not None:
            self.minute = minute

        if second is not None:
            self.second = second

        if nanosecond is not None:
            self.nanosecond = nanosecond


class LocalDateTime(EObject, metaclass=MetaEClass):

    date = EReference(ordered=True, unique=True, containment=True, derived=False)
    time = EReference(ordered=True, unique=True, containment=True, derived=False)

    def __init__(self, *, date=None, time=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if date is not None:
            self.date = date

        if time is not None:
            self.time = time


class Point(EObject, metaclass=MetaEClass):

    line = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)
    column = EAttribute(eType=EInt, unique=True, derived=False, changeable=True)

    def __init__(self, *, line=None, column=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if line is not None:
            self.line = line

        if column is not None:
            self.column = column


class Position(EObject, metaclass=MetaEClass):

    start = EReference(ordered=True, unique=True, containment=True, derived=False)
    end = EReference(ordered=True, unique=True, containment=True, derived=False)

    def __init__(self, *, start=None, end=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if start is not None:
            self.start = start

        if end is not None:
            self.end = end


@abstract
class Origin(EObject, metaclass=MetaEClass):

    def __init__(self):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


@abstract
class Destination(EObject, metaclass=MetaEClass):

    def __init__(self):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


class Statement(EObject, metaclass=MetaEClass):

    def __init__(self):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


class Expression(EObject, metaclass=MetaEClass):

    def __init__(self):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


class EntityDeclaration(EObject, metaclass=MetaEClass):

    def __init__(self):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


class Issue(EObject, metaclass=MetaEClass):

    type = EAttribute(eType=IssueType, unique=True, derived=False, changeable=True)
    message = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    severity = EAttribute(eType=IssueSeverity, unique=True, derived=False, changeable=True)
    position = EReference(ordered=True, unique=True, containment=True, derived=False)

    def __init__(self, *, type=None, message=None, severity=None, position=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if type is not None:
            self.type = type

        if message is not None:
            self.message = message

        if severity is not None:
            self.severity = severity

        if position is not None:
            self.position = position


class PossiblyNamed(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)

    def __init__(self, *, name=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name


class ReferenceByName(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
    referenced = EReference(ordered=True, unique=True, containment=False, derived=False)

    def __init__(self, *, name=None, referenced=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if referenced is not None:
            self.referenced = referenced


class Result(EObject, metaclass=MetaEClass):

    root = EReference(ordered=True, unique=True, containment=True, derived=False)
    issues = EReference(ordered=True, unique=True, containment=True, derived=False, upper=-1)

    def __init__(self, *, root=None, issues=None):
        # if kwargs:
        #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if root is not None:
            self.root = root

        if issues:
            self.issues.extend(issues)


class NodeDestination(Destination):

    node = EReference(ordered=True, unique=True, containment=False, derived=False)

    def __init__(self, *, node=None, **kwargs):

        super().__init__(**kwargs)

        if node is not None:
            self.node = node


class TextFileDestination(Destination):

    position = EReference(ordered=True, unique=True, containment=True, derived=False)

    def __init__(self, *, position=None, **kwargs):

        super().__init__(**kwargs)

        if position is not None:
            self.position = position


@abstract
class ASTNode(Origin):

    position = EReference(ordered=True, unique=True, containment=True, derived=False)
    origin = EReference(ordered=True, unique=True, containment=False, derived=False)
    destination = EReference(ordered=True, unique=True, containment=True, derived=False)

    def __init__(self, *, position=None, origin=None, destination=None, **kwargs):

        super().__init__(**kwargs)

        if position is not None:
            self.position = position

        if origin is not None:
            self.origin = origin

        if destination is not None:
            self.destination = destination


class Named(PossiblyNamed):

    name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)

    def __init__(self, *, name=None, **kwargs):

        super().__init__(**kwargs)

        if name is not None:
            self.name = name
